"""
Handles P2P connections.
All networking functions are ultimately done through
this class.
"""

import hashlib
import signal
import zlib
from ast import literal_eval

from .nat_pmp import NatPMP
from .rendezvous_client import *
from .unl import UNL
from .upnp import *

# How many times a single message can be retransmitted.
max_retransmissions = 1

# Minimum time that must pass between retransmissions.
min_retransmit_interval = 5

# A theoretical time for a message to propagate across the network.
propagation_delay = 5

# A table of message hashes for received messages.
seen_messages = {}

# How often to get new DHT messages.
dht_msg_interval = 5

# How often to re-bootstrap.
rendezvous_interval = 30 * 60

# How often to re-advertise node.
# Update bootstrapping server every 24 hours.
advertise_interval = 60 * 60 * 12

# Time that must elapse between accepting simultaneous opens.
sim_open_interval = 2

# Bootstrapping + TCP hole punching server.
rendezvous_servers = [
    {
        "addr": "185.61.148.22",
        "port": 8000
    },
    {
        "addr": "185.86.149.128",
        "port": 8000
    }
]

# Web server running script to test port forwarding.
# And get WAN IP address.
forwarding_servers = [
    {
        "addr": "185.86.149.128",
        "port": 80,
        "url": "/net.php"
    },
    {
        "addr": "185.61.148.22",
        "port": 80,
        "url": "/net.php"
    }
]

# Debug logging.
logging.basicConfig()
log = logging.getLogger(__name__)


def is_msg_old(msg, record_seen=0):
    if type(msg) == str:
        msg = msg.encode("ascii")

    response_hash = hashlib.sha256(msg).hexdigest()
    if response_hash in seen_messages:
        seen = seen_messages[response_hash]
        elapsed = int(time.time()) - seen["last"]
        if elapsed < min_retransmit_interval:
            return 1

        if seen["times"] >= max_retransmissions:
            return 1

    if record_seen:
        record_msg_hash(msg)

    return 0


def record_msg_hash(msg):
    if type(msg) == str:
        msg = msg.encode("ascii")
    response_hash = hashlib.sha256(msg).hexdigest()

    if not is_msg_old(msg):
        timestamp = int(time.time())
        if response_hash in seen_messages:
            seen = seen_messages[response_hash]
            seen["times"] += 1
            seen["last"] = timestamp
        else:
            seen_messages[response_hash] = {
                "times": 1,
                "last": timestamp
            }

        return 1
    else:
        return 0


def clear_seen_messages():
    global seen_messages
    seen_messages = {}


class Net:
    def __init__(self, net_type="p2p", nat_type="unknown", node_type="unknown",
                 max_outbound=10, max_inbound=10, passive_bind="0.0.0.0",
                 passive_port=50500, interface="default", wan_ip=None,
                 dht_node=None, error_log_path="error.log", debug=0,
                 sys_clock=None, servers=None):
        # List of outbound connections (from us, to another node.)
        self.outbound = []

        # List of inbound connections (to us, from another node.)
        self.inbound = []

        # Socket to receive inbound connections on.
        self.passive = None

        # Type of node: simultaneous, active, passive.
        self.node_type = node_type

        # NAT type: preserving, delta, reuse, random.
        self.nat_type = nat_type

        # Address to listen() on for inbound cons.
        self.passive_bind = passive_bind

        # Port to listen() on for inbound cons.
        self.passive_port = int(passive_port)

        # How many connections can we accept from other nodes?
        self.max_outbound = int(max_outbound)

        # How many connections can we make to other nodes?
        self.max_inbound = int(max_inbound)

        # List of servers to do port forwarding checks.
        self.forwarding_servers = forwarding_servers

        # Unix timestamp of last bootstrap.
        self.last_bootstrap = None

        # Unix timestamp of last DHT direct message.
        self.last_dht_msg = None

        # Unix timestamp of last advertise.
        self.last_advertise = None

        # What interface to make outbound connections from?
        self.interface = interface

        # Skip advertise if we have at least this many inbound connections.
        self.min_connected = 3

        # Unix timestamp of last simultaneous open challenge.
        self.last_passive_sim_open = 0

        # Does this Net instance need to bootstrap?
        self.enable_bootstrap = 1

        # Does this Net instance need to advertise?
        self.enable_advertise = 1

        # Should we try open ports?
        self.enable_forwarding = 1

        # Is simultaneous open enabled?
        self.enable_simultaneous = 1

        # Does this Net instance reject duplicate messages
        # (same hash as previous messages)?
        self.enable_duplicates = 1

        # Where should I store errors?
        self.error_log_path = error_log_path

        # Indicates port forwarding state.
        self.forwarding_type = "manual"

        # Debug mode shows debug messages.
        self.debug = debug

        # Network: p2p or direct.
        self.net_type = net_type

        # Calculate clock skew from NTP.
        self.sys_clock = sys_clock

        # List of rendezvous servers.
        self.rendezvous_servers = servers or rendezvous_servers

        # Rendezvous / boostrapping client.
        self.rendezvous = RendezvousClient(
            self.nat_type, rendezvous_servers=self.rendezvous_servers,
            interface=self.interface,
            sys_clock=self.sys_clock
        )

        # DHT node for receiving direct messages from other nodes.
        self.dht_node = dht_node

        # DHT messages received from DHT.
        self.dht_messages = []

        # Subscribes to certain messages from DHT.
        # Todo: move status messages to file transfer client
        def build_dht_msg_handler():
            def dht_msg_handler(node, msg):
                self.debug_print("DHT msg handler in Net")
                valid_needles = [
                    '^REVERSE_CONNECT',
                    '^REVERSE_QUERY',
                    '^REVERSE_ORIGIN',
                    """u?("|')status("|')(:|,)\s+u?("|')SYN("|')""",
                    """u?("|')status("|')(:|,)\s+u?("|')SYN-ACK("|')""",
                    """u?("|')status("|')(:|,)\s+u?("|')ACK("|')""",
                    """u?("|')status("|')(:|,)\s+u?("|')RST("|')""",
                ]

                # Convert zlib packed binary to Python object.
                self.debug_print("In net dht" + str(type(msg)))
                if type(msg) == type(b""):
                    try:
                        msg = literal_eval(zlib.decompress(msg))
                    except:
                        pass

                    # Encode result to unicode for RE checks.
                    """
                    If buffer errors result: enable this.

                    try:
                        if sys.version_info >= (3, 0, 0):
                            if type(msg) == bytes:
                                msg = msg.decode("utf-8")
                        else:
                            if type(msg) == str:
                                msg = unicode(msg)
                    except:
                        return
                    """

                # Check for matches.
                for needle in valid_needles:
                    if re.search(needle, str(msg)) is not None:
                        msg = {
                            u"message": msg,
                            u"source": None
                        }
                        self.dht_messages.append(msg)
                        return

            return dht_msg_handler

        # Add message handler to DHT for our messages.
        self.dht_msg_handler = build_dht_msg_handler()
        if self.dht_node is not None:
            self.dht_node.add_message_handler(self.dht_msg_handler)

        # External IP of this node.
        self.wan_ip = wan_ip or get_wan_ip()

        # Node type details only known after network is start()ed.
        self.unl = None

        # List of connections that still need to respond to
        # our reverse query.
        self.pending_reverse_queries = []

        # Time frame for connection to respond to reverse query.
        self.reverse_query_expiry = 60

        # Enable more than one connection to the same IP.
        self.enable_duplicate_ip_cons = 0

        # Net instances hide their con details to prioritise direct cons.
        if self.net_type == "direct":
            self.disable_bootstrap()
            self.enable_duplicate_ip_cons = 1

        # Set to 1 when self.start() has been called.
        self.is_net_started = 0

    def debug_print(self, msg):
        log.debug(str(msg))

    def disable_duplicates(self):
        self.enable_duplicates = 0

    def disable_bootstrap(self):
        self.enable_bootstrap = 0

    def disable_advertise(self):
        self.enable_advertise = 0

    def disable_simultaneous(self):
        self.enable_simultaneous = 0

    def disable_forwarding(self):
        self.enable_forwarding = 0

    def get_connection_no(self):
        return len(self.outbound) + len(self.inbound)

    # Used to reject duplicate connections.
    def validate_node(self, node_ip, node_port=None, same_nodes=1):
        self.debug_print("Validating: " + node_ip)

        # Is this a valid IP?
        if not is_ip_valid(node_ip) or node_ip == "0.0.0.0":
            self.debug_print("Invalid node ip in validate node")
            return 0

        # Is this a valid port?
        if node_port != 0 and node_port is not None:
            if not is_valid_port(node_port):
                self.debug_print("Invalid node port in validate port")
                return 0

        """
        Don't accept connections from self to passive server
        or connections to already connected nodes.
        """
        if not self.enable_duplicate_ip_cons:
            # Don't connect to ourself.
            if (node_ip == "127.0.0.1" or
                    node_ip == get_lan_ip(self.interface) or
                    node_ip == self.wan_ip):
                self.debug_print("Cannot connect to ourself.")
                return 0

            # No, really: don't connect to ourself.
            if node_ip == self.passive_bind and node_port == self.passive_port:
                self.debug_print("Error connecting to same listen server.")
                return 0

            # Don't connect to same nodes.
            if same_nodes:
                for node in self.outbound + self.inbound:
                    try:
                        addr, port = node["con"].s.getpeername()
                        if node_ip == addr:
                            self.debug_print("Already connected to this node.")
                            return 0
                    except Exception as e:
                        print(e)
                        return 0

        return 1

    # Make an outbound con to a passive or simultaneous node.
    def add_node(self, node_ip, node_port, node_type, timeout=5):
        # Correct type for port.
        node_port = int(node_port)

        # Debug info.
        msg = "Attempting to connect to %s:%s:%s" % (
            node_ip, str(node_port), node_type
        )
        self.debug_print(msg)

        # Already connected to them.
        con = None
        try:
            if not self.enable_duplicate_ip_cons:
                for node in self.outbound + self.inbound:
                    if node_ip == node["ip"]:
                        self.debug_print("Already connected.")
                        con = node["con"]
                        return con

            # Avoid connecting to ourself.
            if not self.validate_node(node_ip, node_port):
                self.debug_print("Validate node failed.")
                return None

            # Make a simultaneous open connection.
            if node_type == "simultaneous" and self.enable_simultaneous:
                # Check they've started net first
                # If they haven't we won't know the NAT details / node type.
                if not self.is_net_started:
                    raise Exception("Make sure to start net before you add"
                                    " node.")

                if self.nat_type in self.rendezvous.predictable_nats:
                    # Attempt to make active simultaneous connection.
                    old_timeout = self.rendezvous.timeout
                    try:
                        self.rendezvous.timeout = timeout
                        self.debug_print("Attempting simultaneous challenge")
                        con = self.rendezvous.simultaneous_challenge(
                            node_ip, node_port, "TCP"
                        )
                    except Exception as e:
                        self.debug_print("sim challenge failed")
                        error = parse_exception(e)
                        self.debug_print(error)
                        log_exception(self.error_log_path, error)
                        return None
                    self.rendezvous.timeout = old_timeout

                    # Record node details and return con.
                    self.rendezvous.simultaneous_cons = []
                    if con is not None:
                        node = {
                            "con": con,
                            "type": "simultaneous",
                            "ip": node_ip,
                            "port": 0
                        }
                        self.outbound.append(node)
                        self.debug_print("SUCCESS")
                    else:
                        self.debug_print("FAILURE")

            # Passive outbound -- easiest to connect to.
            if node_type == "passive":
                try:
                    # Try connect to passive server.
                    con = Sock(node_ip, node_port, blocking=0,
                               timeout=timeout, interface=self.interface)
                    node = {
                        "con": con,
                        "type": "passive",
                        "ip": node_ip,
                        "port": node_port
                    }
                    self.outbound.append(node)
                    self.debug_print("SUCCESS")
                except Exception as e:
                    self.debug_print("FAILURE")
                    error = parse_exception(e)
                    self.debug_print(error)
                    log_exception(self.error_log_path, error)
                    return None

            # Return new connection.
            return con
        finally:
            # Remove undesirable messages from replies.
            # Save message: 0 = no, 1 = yes.
            def filter_msg_check_builder():
                def filter_msg_check(msg):
                    # Allow duplicate replies?
                    record_seen = not self.enable_duplicates

                    # Check if message is old.
                    return not is_msg_old(msg, record_seen)

                return filter_msg_check

            # Patch sock object to reject duplicate replies
            # If it's enabled.
            if con is not None:
                con.reply_filter = filter_msg_check_builder()

    def bootstrap(self):
        """
        When the software is first started, it needs to retrieve
        a list of nodes to connect to the network to. This function
        asks the server for N nodes which consists of at least N
        passive nodes and N simultaneous nodes. The simultaneous
        nodes are prioritized if the node_type for the machine
        running this software is simultaneous, with passive nodes
        being used as a fallback. Otherwise, the node exclusively
        uses passive nodes to bootstrap.

        This algorithm is designed to preserve passive node's
        inbound connection slots.
        """
        # Disable bootstrap.
        if not self.enable_bootstrap:
            return None

        # Avoid raping the rendezvous server.
        t = time.time()
        if self.last_bootstrap is not None:
            if t - self.last_bootstrap <= rendezvous_interval:
                self.debug_print("Bootstrapped recently")
                return None
        self.last_bootstrap = t
        self.debug_print("Searching for nodes to connect to.")

        try:
            connection_slots = self.max_outbound - (len(self.outbound))
            if connection_slots > 0:
                # Connect to rendezvous server.
                rendezvous_con = self.rendezvous.server_connect()

                # Retrieve random nodes to bootstrap with.
                rendezvous_con.send_line("BOOTSTRAP " +
                                         str(self.max_outbound * 2))
                choices = rendezvous_con.recv_line(timeout=2)
                if choices == "NODES EMPTY":
                    rendezvous_con.close()
                    self.debug_print("Node list is empty.")
                    return self
                else:
                    self.debug_print("Found node list.")

                # Parse node list.
                choices = re.findall("(?:(p|s)[:]([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+)[:]([0-9]+))+\s?", choices)
                rendezvous_con.s.close()

                # Attempt to make active simultaneous connections.
                passive_nodes = []
                for node in choices:
                    # Out of connection slots.
                    if not connection_slots:
                        break

                    # Add to list of passive nodes.
                    node_type, node_ip, node_port = node
                    self.debug_print(str(node))
                    if node_type == "p":
                        passive_nodes.append(node)

                # Use passive to make up the remaining cons.
                i = 0
                while i < len(passive_nodes) and connection_slots > 0:
                    node_type, node_ip, node_port = passive_nodes[i]
                    con = self.add_node(node_ip, node_port, "passive")
                    if con is not None:
                        connection_slots -= 1
                        self.debug_print("Con successful.")
                    else:
                        self.debug_print("Con failed.")

                    i += 1

        except Exception as e:
            self.debug_print("Unknown error in bootstrap()")
            error = parse_exception(e)
            log_exception(self.error_log_path, error)

        return self

    def advertise(self):
        """
        This function tells the rendezvous server that our node is ready to
        accept connections from other nodes on the P2P network that run the
        bootstrap function. It's only used when net_type == p2p
        """

        # Advertise is disabled.
        if not self.enable_advertise:
            self.debug_print("Advertise is disbled!")
            return None

        # Direct net server is reserved for direct connections only.
        if self.net_type == "direct" and self.node_type == "passive":
            return None

        # Net isn't started!.
        if not self.is_net_started:
            raise Exception("Please call start() before you call advertise()")

        # Avoid raping the rendezvous server with excessive requests.
        t = time.time()
        if self.last_advertise is not None:
            if t - self.last_advertise <= advertise_interval:
                return None

            if len(self.inbound) >= self.min_connected:
                return None

        self.last_advertise = t

        # Tell rendezvous server to list us.
        try:
            # We're a passive node.
            if self.node_type == "passive" and\
                            self.passive_port is not None and\
                            self.enable_advertise:
                self.rendezvous.passive_listen(self.passive_port,
                                               self.max_inbound)

            """
            Simultaneous open is only used as a fail-safe for connections to
            nodes on the direct_net and only direct_net can list itself as
            simultaneous so its safe to leave this enabled.
            """
            if self.node_type == "simultaneous":
                self.rendezvous.simultaneous_listen()
        except Exception as e:
            error = parse_exception(e)
            log_exception(self.error_log_path, error)

        return self

    def determine_node(self):
        """
        Determines the type of node based on a combination of forwarding
        reachability and NAT type.
        """

        # Manually set node_type as simultaneous.
        if self.node_type == "simultaneous":
            if self.nat_type != "unknown":
                return "simultaneous"

        # Get IP of binding interface.
        unspecific_bind = ["0.0.0.0", "127.0.0.1", "localhost"]
        if self.passive_bind in unspecific_bind:
            lan_ip = get_lan_ip(self.interface)
        else:
            lan_ip = self.passive_bind

        # Passive node checks.
        if lan_ip is not None \
                and self.passive_port is not None and self.enable_forwarding:
            self.debug_print("Checking if port is forwarded.")

            # Check port isn't already forwarded.
            if is_port_forwarded(lan_ip, self.passive_port, "TCP",
                                 self.forwarding_servers):
                msg = "Port already forwarded. Skipping NAT traversal."
                self.debug_print(msg)

                self.forwarding_type = "forwarded"
                return "passive"
            else:
                self.debug_print("Port is not already forwarded.")

            # Most routers.
            try:
                self.debug_print("Trying UPnP")

                UPnP(self.interface).forward_port("TCP", self.passive_port,
                                                  lan_ip)

                if is_port_forwarded(lan_ip, self.passive_port, "TCP",
                                     self.forwarding_servers):
                    self.forwarding_type = "UPnP"
                    self.debug_print("Forwarded port with UPnP.")
                else:
                    self.debug_print("UPnP failed to forward port.")

            except Exception as e:
                # Log exception.
                error = parse_exception(e)
                log_exception(self.error_log_path, error)
                self.debug_print("UPnP failed to forward port.")

                # Apple devices.
                try:
                    self.debug_print("Trying NATPMP.")
                    NatPMP(self.interface).forward_port("TCP",
                                                        self.passive_port,
                                                        lan_ip)
                    if is_port_forwarded(lan_ip, self.passive_port, "TCP",
                                         self.forwarding_servers):
                        self.forwarding_type = "NATPMP"
                        self.debug_print("Port forwarded with NATPMP.")
                    else:
                        self.debug_print("Failed to forward port with NATPMP.")
                        self.debug_print("Falling back on TCP hole punching or"
                                         " proxying.")
                except Exception as e:
                    # Log exception
                    error = parse_exception(e)
                    log_exception(self.error_log_path, error)
                    self.debug_print("Failed to forward port with NATPMP.")

            # Check it worked.
            if self.forwarding_type != "manual":
                return "passive"

        # Fail-safe node types.
        if self.nat_type != "unknown":
            return "simultaneous"
        else:
            return "active"

    # Receive inbound connections.
    def start_passive_server(self):
        self.passive = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.passive.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.passive.bind((self.passive_bind, self.passive_port))
        self.passive.listen(self.max_inbound)

        # Check bound local port.
        if not self.passive_port:
            self.passive_port = self.passive.getsockname()[1]

    def start(self):
        """
        This function determines node and NAT type, saves connectivity details,
        and starts any needed servers to be a part of the network. This is
        usually the first function called after initialising the Net class.
        """

        self.debug_print("Starting networking.")
        self.debug_print("Make sure to iterate over replies if you need"
                         " connection alive management!")

        # Register a cnt + c handler
        signal.signal(signal.SIGINT, self.stop)

        # Save WAN IP.
        self.debug_print("WAN IP = " + str(self.wan_ip))

        # Check rendezvous server is up.
        try:
            rendezvous_con = self.rendezvous.server_connect()
            rendezvous_con.close()
        except:
            raise Exception("Unable to connect to rendezvous server.")

        # Started no matter what
        # since LAN connections are always possible.
        self.start_passive_server()

        # Determine NAT type.
        if self.nat_type == "unknown":
            self.debug_print("Determining NAT type.")
            nat_type = self.rendezvous.determine_nat()
            if nat_type is not None and nat_type != "unknown":
                self.nat_type = nat_type
                self.rendezvous.nat_type = nat_type
                self.debug_print("NAT type = " + nat_type)
            else:
                self.debug_print("Unable to determine NAT type.")

        # Check NAT type if node is simultaneous
        # is manually specified.
        if self.node_type == "simultaneous":
            if self.nat_type not in self.rendezvous.predictable_nats:
                self.debug_print("Manual setting of simultanous specified but"
                                 " ignored since NAT does not support it.")
                self.node_type = "active"
        else:
            # Determine node type.
            self.debug_print("Determining node type.")

            # No checks for manually specifying passive
            # (there probably should be.)
            if self.node_type == "unknown":
                self.node_type = self.determine_node()

        # Prevent P2P nodes from running as simultaneous.
        if self.net_type == "p2p":
            """
            TCP hole punching is reserved specifically for direct networks
            (a net object reserved for receiving direct connections
            -- p2p is for connecting to the main network. The reason for this
            is you can't do multiple TCP hole punches at the same time so
            reserved for direct network where it's most needed.
            """
            if self.node_type == "simultaneous":
                self.debug_print("Simultaneous is not allowed for P2P")
                self.node_type = "active"
                self.disable_simultaneous()

        self.debug_print("Node type = " + self.node_type)

        # Close stray cons from determine_node() tests.
        self.close_cons()

        # Set net started status.
        self.is_net_started = 1

        # Initialise our UNL details.
        self.unl = UNL(
            net=self,
            dht_node=self.dht_node,
            wan_ip=self.wan_ip
        )

        # Nestled calls.
        return self

    def stop(self, signum=None, frame=None):
        self.debug_print("Stopping networking.")

        if self.passive is not None:
            try:
                self.passive.shutdown(1)
            except:
                pass
            self.passive.close()
            self.passive = None

        if self.last_advertise is not None:
            self.rendezvous.leave_fight()

        """
        Just let the threads timeout by themselves.
        Otherwise mutex deadlocks could occur.
        for unl_thread in self.unl.unl_threads:
            unl_thread.exit()
        """

        for con in self:
            con.close()

        if signum is not None:
            raise Exception("Process was interrupted.")

    # Return a connection that matches a remote UNL.
    def con_by_unl(self, unl, cons=None):
        if cons is None:
            cons = self.outbound + self.inbound
        for con in cons:
            if not isinstance(con, Sock):
                con = con["con"]

            if con.unl is not None:
                self.debug_print("CMP")
                self.debug_print(unl)
                self.debug_print(con.unl)
                if unl == con.unl:
                    # Connection not ready.
                    if con.nonce is None and self.net_type == "direct":
                        continue

                    return con
            else:
                self.debug_print("\a")
                self.debug_print("Con UNL is None (in con by unl)")
                self.debug_print(cons)

        return None

    # Return a connection by its IP.
    def con_by_ip(self, ip):
        for node in self.outbound + self.inbound:
            # Used to block UNLs until nonces are received.
            # Otherwise they might try do I/O and ruin their protocols.
            if self.net_type == "direct":
                if node["con"].nonce is None and self.net_type == "direct":
                    continue

            if node["ip"] == ip:
                return node["con"]

        return None

    def generate_con_id(self, nonce, their_wan_ip, our_wan_ip):
        # Convert WAN IPs to bytes.
        if sys.version_info >= (3, 0, 0):
            if type(their_wan_ip) == str:
                their_wan_ip = their_wan_ip.encode("ascii")

                if type(our_wan_ip) == str:
                    our_wan_ip = our_wan_ip.encode("ascii")
        else:
            if type(their_wan_ip) == unicode:
                their_wan_ip = str(their_wan_ip)

            if type(our_wan_ip) == our_wan_ip:
                our_wan_ip = str(our_wan_ip)

        # Hash WAN IPs to make them the same length.
        their_wan_ip = hashlib.sha256(their_wan_ip).hexdigest().encode("ascii")
        our_wan_ip = hashlib.sha256(our_wan_ip).hexdigest().encode("ascii")

        # Derive fingerprint.
        int_their_wan_ip = int(their_wan_ip, 16)
        int_our_wan_ip = int(our_wan_ip, 16)
        if int_our_wan_ip > int_their_wan_ip:
            fingerprint = hashlib.sha256(our_wan_ip + their_wan_ip)
        else:
            # If both are the same the order doesn't matter.
            fingerprint = hashlib.sha256(their_wan_ip + our_wan_ip)
        fingerprint = fingerprint.hexdigest().encode("ascii")

        # Convert nonce to bytes.
        if sys.version_info >= (3, 0, 0):
            if type(nonce) == str:
                nonce = nonce.encode("ascii")
        else:
            if type(nonce) == unicode:
                nonce = str(nonce)

        # Generate con ID.
        con_id = hashlib.sha256(nonce + fingerprint).hexdigest()

        # Convert to unicode.
        if sys.version_info >= (3, 0, 0):
            if type(con_id) == bytes:
                con_id = con_id.decode("utf-8")
        else:
            if type(con_id) == str:
                con_id = unicode(con_id)

        # Return con ID.
        return con_id

    def con_by_id(self, expected_id):
        for node in self.outbound + self.inbound:
            # Nothing to test.
            if node["con"].nonce is None and self.net_type == "direct":
                self.debug_print("Nonce not set")
                continue

            # Generate con_id from con.
            try:
                their_wan_ip, junk = node["con"].s.getpeername()
            except:
                continue
            if is_ip_private(their_wan_ip):
                our_wan_ip = get_lan_ip(self.interface)
            else:
                our_wan_ip = self.wan_ip
            found_id = self.generate_con_id(
                node["con"].nonce,
                their_wan_ip,
                our_wan_ip
            )

            # Check result.
            if found_id == expected_id:
                return node["con"]

        return None

    # Send a message to all currently established connections.
    def broadcast(self, msg, source_con=None):
        for node in self.outbound + self.inbound:
            if node["con"] != source_con:
                node["con"].send_line(msg)

    def close_cons(self):
        # Close all connections.
        for node in self.inbound + self.outbound:
            node["con"].close()

        # Flush client queue for passive server.
        if self.node_type == "passive" and self.passive is not None:
            self.passive.close()
            self.start_passive_server()

        # Start from scratch.
        self.inbound = []
        self.outbound = []

    def synchronize(self):
        # Clean up dead connections.
        for node_list_name in ["self.inbound", "self.outbound"]:
            node_list = eval(node_list_name)[:]
            for node in node_list:
                if not node["con"].connected:
                    self.debug_print("\a")
                    self.debug_print("Removing disconnected: " + str(node))
                    eval(node_list_name).remove(node)

        # Timeout connections that haven't responded to reverse query.
        old_reverse_queries = []
        for reverse_query in self.pending_reverse_queries:
            duration = time.time() - reverse_query["timestamp"]
            if duration >= self.reverse_query_expiry:
                reverse_query["con"].close()
                old_reverse_queries.append(reverse_query)

        # Remove old reverse queries.
        for reverse_query in old_reverse_queries:
            self.pending_reverse_queries.remove(reverse_query)

        # Get connection nonce (for building IDs.)
        if self.net_type == "direct":
            for node in self.inbound + self.outbound:
                if node["con"].nonce is not None:
                    continue

                # Receive nonce part.
                if len(node["con"].nonce_buf) < 64:
                    assert(node["con"].blocking != 1)
                    remaining = 64 - len(node["con"].nonce_buf)
                    nonce_part = node["con"].recv(remaining)
                    if len(nonce_part):
                        node["con"].nonce_buf += nonce_part

                # Set nonce.
                if len(node["con"].nonce_buf) == 64:
                    node["con"].nonce = node["con"].nonce_buf

        # Check for reverse connect requests.
        if self.dht_node is not None and self.net_type == "direct":
            # Don't do this every synch cycle.
            t = time.time()
            skip_dht_check = 0
            if self.last_dht_msg is not None:
                if t - self.last_dht_msg > dht_msg_interval:
                    skip_dht_check = 1

            if not skip_dht_check and len(self.dht_messages):
                processed = []
                for dht_response in self.dht_messages:
                    # Found reverse connect request.
                    msg = str(dht_response["message"])
                    if re.match("^REVERSE_CONNECT:[a-zA-Z0-9+/-=_\s]+:[a-fA-F0-9]{64}$", msg) is not None:
                        # Process message.
                        self.debug_print(str(msg))
                        call, their_unl, nonce = msg.split(":")
                        their_unl = UNL(value=their_unl).deconstruct()
                        our_unl = UNL(value=self.unl.value).deconstruct()
                        node_id = their_unl["node_id"]

                        # Are we already connected.
                        is_connected = False
                        if nonce == "0" * 64:
                            # Use LAN IPs.
                            their_ip = their_unl["wan_ip"]
                            our_ip = our_unl["wan_ip"]
                            if their_ip == our_ip:
                                their_ip = their_unl["lan_ip"]
                                our_ip = our_unl["lan_ip"]

                            # Get con ID.
                            con_id = self.generate_con_id(
                                nonce,
                                their_ip,
                                our_ip
                            )

                            # Find con if it exists.
                            if self.con_by_id(con_id) is not None:
                                is_connected = True
                        else:
                            if self.con_by_unl(their_unl) is not None:
                                is_connected = True

                        # Skip if already connected.
                        if is_connected:
                            processed.append(dht_response)
                            continue

                        # Ask if the source sent it.
                        def success_builder():
                            def success(con):
                                # Indicate status.
                                self.debug_print("Received reverse connect"
                                                 " notice")
                                self.debug_print(nonce)

                                # Did you send this?
                                query = "REVERSE_QUERY:" + self.unl.value
                                self.dht_node.repeat_relay_message(node_id,
                                                                   query)

                                # Record pending query state.
                                query = {
                                    "unl": their_unl["value"],
                                    "con": con,
                                    "timestamp": time.time()
                                }
                                self.pending_reverse_queries.append(query)

                            return success

                        self.debug_print("Attempting to do reverse connect")
                        self.unl.connect(their_unl["value"],
                                         {"success": success_builder()},
                                         nonce=nonce)

                        processed.append(dht_response)

                    # Found reverse query (did you make this?)
                    elif re.match("^REVERSE_QUERY:[a-zA-Z0-9+/-=_\s]+$", msg)\
                            is not None:
                        # Process message.
                        self.debug_print("Received reverse query")
                        call, their_unl = msg.split(":")
                        their_unl = UNL(value=their_unl).deconstruct()
                        node_id = their_unl["node_id"]

                        # Do we know about this?
                        if their_unl["value"] not in \
                                self.unl.pending_reverse_con:
                            self.debug_print(their_unl)
                            self.debug_print(str(self.unl.pending_reverse_con))
                            self.debug_print("oops, we don't know about this"
                                             " reverse query!")
                            processed.append(dht_response)
                            continue
                        else:
                            self.unl.pending_reverse_con.remove(
                                    their_unl["value"])

                        # Send query.
                        query = "REVERSE_ORIGIN:" + self.unl.value
                        self.dht_node.repeat_relay_message(node_id, query)

                        processed.append(dht_response)

                    # Found reverse origin (yes I made this.)
                    elif re.match("^REVERSE_ORIGIN:[a-zA-Z0-9+/-=_\s]+$", msg) \
                            is not None:
                        self.debug_print("Received reverse origin")
                        for reverse_query in self.pending_reverse_queries:
                            pattern = "^REVERSE_ORIGIN:" + reverse_query["unl"]
                            pattern += "$"
                            if re.match(pattern, msg) is not None:
                                self.debug_print("Removing pending reverse"
                                                 " query: success!")
                                self.pending_reverse_queries.remove(
                                        reverse_query)
                                processed.append(dht_response)

                # Remove processed messages.
                for msg in processed:
                    self.debug_print(msg)
                    self.dht_messages.remove(msg)

            self.last_dht_msg = t

        # Accept inbound connections.
        if len(self.inbound) < self.max_inbound:
            # Accept new passive inbound connections.
            if self.passive is not None:
                r, w, e = select.select([self.passive], [], [], 0)
                for s in r:
                    if s == self.passive:
                        # Accept a new con from the listen queue.
                        client, address = self.passive.accept()
                        con = Sock(blocking=0)
                        con.set_sock(client)
                        node_ip, node_port = con.s.getpeername()

                        # Reject duplicate connections.
                        if self.validate_node(node_ip, node_port):
                            try:
                                node = {
                                    "type": "accept",
                                    "con": con,
                                    "ip": con.s.getpeername()[0],
                                    "port": con.s.getpeername()[1],
                                }
                                self.inbound.append(node)
                                self.debug_print(
                                        "Accepted new passive connection: " +
                                        str(node))
                            except:
                                log.debug("con.s.get")
                        else:
                            self.debug_print("Validation failure")
                            con.close()

            # Accept new passive simultaneous connections.
            if self.node_type == "simultaneous":
                """
                This is basically the code that passive simultaneous
                nodes periodically call to parse any responses from the
                Rendezvous Server which should hopefully be new
                requests to initiate hole punching from active
                simultaneous nodes.

                If a challenge comes in, the passive simultaneous
                node accepts the challenge by giving details to the
                server for the challenging node (active simultaneous)
                to complete the simultaneous open.
                """

                # try:
                t = time.time()
                if self.rendezvous.server_con is not None:
                    for reply in self.rendezvous.server_con:
                        # Reconnect.
                        if re.match("^RECONNECT$", reply) is not None:
                            if self.enable_advertise:
                                self.rendezvous.simultaneous_listen()
                            continue

                        # Find any challenges.
                        # CHALLENGE 192.168.0.1 50184 50185 50186 50187 TCP
                        parts = re.findall("^CHALLENGE ([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+) ((?:[0-9]+\s?)+) (TCP|UDP)$", reply)
                        if not len(parts):
                            continue
                        (candidate_ip, candidate_predictions, candidate_proto)\
                            = parts[0]
                        self.debug_print("Found challenge")
                        self.debug_print(parts[0])

                        # Already connected.
                        if not self.validate_node(candidate_ip):
                            self.debug_print("validation failed")
                            continue

                        # Last meeting was too recent.
                        if t - self.last_passive_sim_open < sim_open_interval:
                            continue

                        # Accept challenge.
                        if self.sys_clock is not None:
                            origin_ntp = self.sys_clock.time()
                        else:
                            origin_ntp = get_ntp()
                        if origin_ntp is None:
                            continue
                        msg = "ACCEPT %s %s TCP %s" % (
                            candidate_ip,
                            self.rendezvous.predictions,
                            str(origin_ntp)
                        )
                        ret = self.rendezvous.server_con.send_line(msg)
                        if not ret:
                            continue

                        """
                        Adding threading here doesn't work because Python's
                        fake threads and the act of starting a thread ruins
                        the timing between code synchronisation - especially
                        code running on the same host or in a LAN. Will
                        compensate by reducing the NTP delay to have the
                        meetings occur faster and setting a limit for meetings
                        to occur within the same period.
                        """
                        # Walk to fight and return holes made.
                        self.last_passive_sim_open = t
                        con = self.rendezvous.attend_fight(
                            self.rendezvous.mappings, candidate_ip,
                            candidate_predictions, origin_ntp
                        )
                        if con is not None:
                            try:
                                node = {
                                    "type": "simultaneous",
                                    "con": con,
                                    "ip": con.s.getpeername()[0],
                                    "port": con.s.getpeername()[1],
                                }
                                self.inbound.append(node)
                            except:
                                log.debug(str(e))
                                pass

                        # Create new predictions ready to accept next client.
                        self.rendezvous.simultaneous_cons = []
                        if self.enable_advertise:
                            self.rendezvous.simultaneous_listen()

        # QUIT - remove us from bootstrapping server.
        if len(self.inbound) == self.max_inbound:
            try:
                # Remove advertise.
                self.rendezvous.leave_fight()
            except:
                pass

        # Bootstrap again if needed.
        self.bootstrap()

        # Relist node again if noded.
        self.advertise()

    """
    These functions here make the class behave like a list. The
    list is a collection of connections (inbound) + (outbound.)
    Every iteration also has the bonus of reaping dead connections,
    making new ones (if needed), and accepting connections
    """
    def __len__(self):
        self.synchronize()
        return len(self.inbound) + len(self.outbound)

    def __iter__(self):
        # Process connections.
        self.synchronize()

        # Copy all connections to single buffer.
        cons = []
        for node in self.inbound + self.outbound:
            if node["con"].nonce is None:
                if self.net_type == "direct":
                    continue

            cons.append(node["con"])

        # Return all cons.
        return iter(cons)

if __name__ == "__main__":
    """
    net = Net(debug=1)
    net.disable_bootstrap()
    net.disable_advertise()
    net.disable_forwarding()
    net.start()
    print(net.unl.value)
    print(net.unl.deconstruct(net.unl.value))

    while 1:
        time.sleep(0.5)

    # Test simultaneous open.
    p2p_net = Net(debug=1, nat_type="preserving", node_type="simultaneous")
    p2p_net.start()
    p2p_net.disable_advertise()
    p2p_net.disable_bootstrap()
    # p2p_net.add_node("192.187.97.131", 0, "simultaneous") # Behind NAT

    def success_notify(con):
        print("SUCCESS THREADING.")

    #Test UNL
    events = {
        "success": success_notify
    }

    while 1:
        time.sleep(0.5)

    exit()

    #P2P network example.
    p2p_net = Net(debug=1)
    p2p_net.start()
    p2p_net.bootstrap()
    p2p_net.advertise()

    #Event loop.
    while 1:
        for con in p2p_net:
            for reply in con:
                print(reply)


            Excluses con from broadcast since we got this message from them

            p2p_net.broadcast("Something.", con)

        time.sleep(0.5)

    #Direct network example.
    dht_node = DHT()
    direct_net = Net(dht_node=dht_node, debug=1)
    direct_net.start()

    #Connect to some UNL.
    def success(con):
        con.send_line("Thanks.")

    #Note: this isn't a valid UNL.
    #To get your UNL do: direct_net.unl.value.
    direct_net.unl.connect("Some guys UNL...", {"success": success})
    """



import platform

if platform.system() == "Windows":
    import ctypes
    from ctypes import windll
    from ctypes import wintypes

    kernel32 = windll.kernel32
    iphlpapi = windll.iphlpapi
    Ws2_32 = windll.Ws2_32
    ERROR_INSUFFICIENT_BUFFER = 122
    ERROR_NO_DATA = 232

    class Win32MIBIPFORWARDROW(ctypes.Structure):
        _fields_ = [
            ('dwForwardDest', wintypes.DWORD),
            ('dwForwardMask', wintypes.DWORD),
            ('dwForwardPolicy', wintypes.DWORD),
            ('dwForwardNextHop', wintypes.DWORD),
            ('dwForwardIfIndex', wintypes.DWORD),
            ('dwForwardType', wintypes.DWORD),
            ('dwForwardProto', wintypes.DWORD),
            ('dwForwardAge', wintypes.DWORD),
            ('dwForwardNextHopAS', wintypes.DWORD),
            ('dwForwardMetric1', wintypes.DWORD),
            ('dwForwardMetric2', wintypes.DWORD),
            ('dwForwardMetric3', wintypes.DWORD),
            ('dwForwardMetric4', wintypes.DWORD),
            ('dwForwardMetric5', wintypes.DWORD)
        ]

    class Win32MIBIPFORWARDTABLE(ctypes.Structure):
        _fields_ = [
            ('dwNumEntries', wintypes.DWORD),
            ('table', Win32MIBIPFORWARDROW * 1)
        ]

    kernel32.GetProcessHeap.argtypes = []
    kernel32.GetProcessHeap.restype = wintypes.HANDLE

    # Note: wintypes.ULONG must be replaced with a 64 bit variable on x64
    kernel32.HeapAlloc.argtypes = [wintypes.HANDLE, wintypes.DWORD,
                                   wintypes.ULONG]
    kernel32.HeapAlloc.restype = wintypes.LPVOID

    kernel32.HeapFree.argtypes = [wintypes.HANDLE, wintypes.DWORD,
                                  wintypes.LPVOID]
    kernel32.HeapFree.restype = wintypes.BOOL

    iphlpapi.GetIpForwardTable.argtypes = [
        ctypes.POINTER(Win32MIBIPFORWARDTABLE),
        ctypes.POINTER(wintypes.ULONG),
        wintypes.BOOL]
    iphlpapi.GetIpForwardTable.restype = wintypes.DWORD

    Ws2_32.inet_ntoa.restype = ctypes.c_char_p

    def get_ipv4_routing_table():
        routing_table = []

        heap = kernel32.GetProcessHeap()

        size = wintypes.ULONG(ctypes.sizeof(Win32MIBIPFORWARDTABLE))
        p = kernel32.HeapAlloc(heap, 0, size)
        if not p:
            raise Exception('Unable to allocate memory for the IP forward '
                            'table')
        p_forward_table = ctypes.cast(
            p, ctypes.POINTER(Win32MIBIPFORWARDTABLE))

        try:
            err = iphlpapi.GetIpForwardTable(p_forward_table,
                                             ctypes.byref(size), 0)
            if err == ERROR_INSUFFICIENT_BUFFER:
                kernel32.HeapFree(heap, 0, p_forward_table)
                p = kernel32.HeapAlloc(heap, 0, size)
                if not p:
                    raise Exception('Unable to allocate memory for the IP '
                                    'forward table')
                p_forward_table = ctypes.cast(
                    p, ctypes.POINTER(Win32MIBIPFORWARDTABLE))

            err = iphlpapi.GetIpForwardTable(p_forward_table,
                                             ctypes.byref(size), 0)
            if err != ERROR_NO_DATA:
                if err:
                    raise Exception('Unable to get IP forward table. '
                                    'Error: %s' % err)

                forward_table = p_forward_table.contents
                table = ctypes.cast(
                    ctypes.addressof(forward_table.table),
                    ctypes.POINTER(Win32MIBIPFORWARDROW *
                                   forward_table.dwNumEntries)).contents

                i = 0
                while i < forward_table.dwNumEntries:
                    row = table[i]
                    routing_table.append((
                        Ws2_32.inet_ntoa(row.dwForwardDest),
                        Ws2_32.inet_ntoa(row.dwForwardMask),
                        Ws2_32.inet_ntoa(row.dwForwardNextHop),
                        row.dwForwardIfIndex,
                        row.dwForwardMetric1))
                    i += 1

            return routing_table
        finally:
            kernel32.HeapFree(heap, 0, p_forward_table)



"""
Implements a custom protocol for sending and receiving
line delineated messages. For blocking sockets,
time-out is required to avoid DoS attacks when talking
to a misbehaving or malicious third party.

The benefit of this class is it makes communication
with the P2P network easy to code without having to
depend on threads and hence on mutexes (which are hard
to use correctly.)

In practice, a connection to a node on the P2P network
would be done using the default options of this class
and the connection would periodically be polled for
replies. The processing of replies would automatically
break once the socket indicated it would block and
to prevent a malicious node from sending replies as
fast as it could - there would be a max message limit
per check period.

Quirks:
* send_line will block until the entire line has been sent even if the socket
  has been set to non-blocking to make things easier. If you need a non-blocking
  way to send a line: use send(). Note that you will have to check for the
  number of bytes sent and resend if needed just like the real send function.
* connect has the same behaviour as above to make things simpler (so will block
  regardless of whether socket is in non-blocking mode or not.) If you want to
  bypass this behaviour you can always connect the socket outside this class
  and then pass it to set_socket.

Otherwise, all functions in this class behave how you would expect them to
(depending on whether you're using non-blocking mode or blocking mode.) It's
assumed that all blocking operations have a timeout by default. This can't be
disabled.

Todo: test various functions under connection exit.
Timeouts are needed for non-blocking too under conditions where you attempt to
send all / recv all.
"""

import errno
import platform
import socket
import ssl
import sys
import time

from pyp2p.lib import get_lan_ip, parse_exception, log_exception

error_log_path = "error.log"


class Sock:
    def __init__(self, addr=None, port=None, blocking=0, timeout=5,
                 interface="default", use_ssl=0, debug=0):
        self.nonce = None
        self.nonce_buf = u""
        self.reply_filter = None
        self.buf = u""
        self.max_buf = 1024 * 1024  # 1 MB.
        self.max_chunks = 1024  # Prevents spamming of multiple short messages.
        self.chunk_size = 1024 * 4
        self.replies = []
        self.blocking = blocking
        self.timeout = timeout
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # self.s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.use_ssl = use_ssl
        self.alive = time.time()
        self.unl = None
        if self.use_ssl:
            self.s = ssl.wrap_socket(self.s)

        self.connected = 0
        self.interface = interface
        self.delimiter = u"\r\n"
        self.debug = debug

        # Set keep alive.
        # self.set_keep_alive(self.s)

        # Connect socket.
        if addr is not None and port is not None:
            # Set a timeout for blocking operations so they don't DoS program.
            # Disabled after connect if non-blocking is set.
            # (Connect is so far always blocking regardless of blocking mode.)
            self.s.settimeout(5)

            self.connect(addr, port)
        else:
            self.set_blocking(self.blocking, self.timeout)

    def debug_print(self, msg):
        msg = "> " + str(msg)
        if self.debug:
            print(msg)

    def set_keep_alive(self, sock, after_idle_sec=5, interval_sec=60,
                       max_fails=5):
        """
        This function instructs the TCP socket to send a heart beat every n
        seconds to detect dead connections. It's the TCP equivalent of the
        IRC ping-pong protocol and allows for better cleanup / detection
        of dead TCP connections.

        It activates after 1 second (after_idle_sec) of idleness, then sends
        a keepalive ping once every 3 seconds(interval_sec), and closes the
        connection after 5 failed ping (max_fails), or 15 seconds
        """

        # OSX
        if platform.system() == "Darwin":
            # scraped from /usr/include, not exported by python's socket module
            TCP_KEEPALIVE = 0x10
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, interval_sec)

        if platform.system() == "Windows":
            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, 10000, 3000))

        if platform.system() == "Linux":
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
                            after_idle_sec)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL,
                            interval_sec)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)

    def set_blocking(self, blocking, timeout=5):
        self.debug_print(self.s is None)
        self.debug_print("TRYING to set blocking")

        if self.s is None:
            return

        self.debug_print("Setting blocking")

        # Update blocking mode.
        self.s.setblocking(blocking)

        # Adjust timeout if needed.
        if blocking:
            if timeout is not None:
                self.s.settimeout(timeout)

        # Update blocking status.
        self.timeout = timeout
        self.blocking = blocking

    def set_sock(self, s):
        self.debug_print("Setting sock")
        self.close()  # Close old socket.
        self.s = s
        self.set_blocking(self.blocking, self.timeout)
        # self.s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # Set keep alive.
        # self.set_keep_alive(self.s)

        # Save addr + port.
        try:
            addr, port = self.s.getpeername()
            self.addr = addr
            self.port = port
            self.connected = 1
        except:
            self.connected = 0

    def reconnect(self):
        if not self.connected:
            if self.addr is not None and self.port is not None:
                try:
                    return self.connect(self.addr, self.port)
                except:
                    self.connected = 0

    # Blocking (regardless of socket mode.)
    def connect(self, addr, port):
        # Save addr and port so socket can be reconnected.
        self.addr = addr
        self.port = port

        # No socket detected.
        if self.s is None:
            self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if self.use_ssl:
                self.s = ssl.wrap_socket(self.s)

        # Make connection from custom interface.
        if self.interface != "default":
            try:
                # Todo: fix this to use static ips from Net
                src_ip = get_lan_ip(self.interface)
                self.s.bind((src_ip, 0))
            except socket.error as e:
                if e.errno != 98:
                    raise e

        try:
            self.debug_print(addr)
            self.debug_print(port)
            self.s.connect((addr, int(port)))
            self.connected = 1
            self.set_blocking(self.blocking, self.timeout)
        except Exception as e:
            self.debug_print("Connect failed")
            error = parse_exception(e)
            self.debug_print(error)
            log_exception(error_log_path, error)
            raise socket.error("Socket connect failed.")

    def close(self):
        self.connected = 0
        self.debug_print("Closing con!!!")

        # Attempt graceful shutdown.
        try:
            try:
                self.s.shutdown(1)
            except:
                pass
            self.s.close()
        except:
            pass

        self.s = None

    def parse_buf(self):
        """
        Since TCP is a stream-orientated protocol, responses aren't guaranteed
        to be complete when they arrive. The buffer stores all the data and
        this function splits the data into replies based on the new line
        delimiter.
        """
        buf_len = len(self.buf)
        replies = []
        reply = u""
        chop = 0
        skip = 0
        i = 0
        for ch in self.buf:
            if skip:
                skip -= 1
                i += 1
                continue

            nxt = i + 1
            if nxt < buf_len:
                if ch == u"\r" and self.buf[nxt] == u"\n":
                    # Append new reply.
                    if reply != u"":
                        replies.append(reply)
                        reply = u""

                    # Truncate the whole buf if chop is out of bounds.
                    chop = nxt + 1
                    skip = 1
                    i += 1
                    continue

            reply += ch
            i += 1

        # Truncate buf.
        if chop:
            self.buf = self.buf[chop:]

        return replies

    # Blocking or non-blocking.
    def get_chunks(self, fixed_limit=None, encoding="unicode"):
        """
        This is the function which handles retrieving new data chunks. It's
        main logic is avoiding a recv call blocking forever and halting
        the program flow. To do this, it manages errors and keeps an eye
        on the buffer to avoid overflows and DoS attacks.

        http://stackoverflow.com/questions/16745409/what-does-pythons-socket-recv-return-for-non-blocking-sockets-if-no-data-is-r
        http://stackoverflow.com/questions/3187565/select-and-ssl-in-python
        """

        # Socket is disconnected.
        if not self.connected:
            return

        # Recv chunks until network buffer is empty.
        repeat = 1
        wait = 0.2
        chunk_no = 0
        max_buf = self.max_buf
        max_chunks = self.max_chunks
        if fixed_limit is not None:
            max_buf = fixed_limit
            max_chunks = fixed_limit

        self.debug_print("Get chunks blocking mode = ")
        self.debug_print(str(self.s.gettimeout()))
        while repeat:
            chunk_size = self.chunk_size
            while True:
                # Don't exceed buffer size.
                buf_len = len(self.buf)
                if buf_len >= max_buf:
                    break
                remaining = max_buf - buf_len
                if remaining < chunk_size:
                    chunk_size = remaining

                # Don't allow non-blocking sockets to be
                # DoSed by multiple small replies.
                if chunk_no >= max_chunks and not self.blocking:
                    break

                try:
                    chunk = self.s.recv(chunk_size)
                except socket.timeout as e:
                    self.debug_print("Get chunks timed out.")
                    self.debug_print(e)

                    # Timeout on blocking sockets.
                    err = e.args[0]
                    self.debug_print(err)
                    if err == "timed out":
                        repeat = 0
                        break
                except ssl.SSLError as e:
                    # Will block on non-blocking SSL sockets.
                    if e.errno == ssl.SSL_ERROR_WANT_READ:
                        self.debug_print("SSL_ERROR_WANT_READ")
                        break
                    else:
                        self.debug_print("Get chunks ssl error")
                        self.close()
                        return
                except socket.error as e:
                    # Will block on nonblocking non-SSL sockets.
                    self.debug_print("Get chunks socket.error")
                    err = e.args[0]
                    self.debug_print(err)
                    if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                        break
                    else:
                        # Connection closed or other problem.
                        self.debug_print("get chunks other closing")
                        self.close()
                        return
                else:
                    if chunk == b"":
                        self.debug_print("Get chunk: b''")
                        self.close()
                        return

                    # Avoid decoding errors.
                    try:
                        if encoding == "unicode":
                            self.buf += chunk.decode("utf-8")
                        else:
                            self.buf += chunk.decode("latin-1")
                        self.alive = time.time()
                    except Exception as e:
                        self.debug_print(e)
                        self.debug_print("Get chunk: can't decode.")
                        chunk_no += 1
                        continue

                    # Otherwise the loop will be endless.
                    if self.blocking:
                        break

                    # Used to avoid DoS of small packets.
                    chunk_no += 1

            # Repeat is already set -- manual skip.
            if not repeat:
                break
            else:
                repeat = 0

            # Block until there's a full reply or there's a timeout.
            if self.blocking:
                if fixed_limit is None and encoding == "unicode":
                    # Partial response.
                    if self.delimiter not in self.buf:
                        repeat = 1
                        time.sleep(wait)

    def reply_callback(self, callback):
        self.reply_callback = callback

    # Called to check for replies and update buffers.
    def update(self):
        self.get_chunks()
        self.replies += self.parse_buf()

        # Execute callbacks on replies.
        if self.reply_filter is not None:
            replies = []
            for reply in self.replies:
                if not self.reply_filter(reply):
                    replies.append(u"")
                else:
                    replies.append(reply)

            self.replies = replies

    # Blocking or non-blocking.
    def send(self, msg, send_all=0, timeout=5):
        # Sanity check.
        assert (len(msg))
        # Not connected.
        if not self.connected:
            return 0

        # Update timeout.
        if timeout != self.timeout and self.blocking:
            self.set_blocking(self.blocking, timeout)

        # Check socket is in correct blocking mode.
        if self.debug:
            blocking = self.s.gettimeout()
            if self.blocking:
                assert (blocking >= 1 or blocking is None)
            else:
                assert (blocking == 0.0)

        try:
            # Convert to bytes Python 2 & 3
            if sys.version_info >= (3, 0, 0):
                if type(msg) == str:
                    msg = msg.encode("ascii")
            else:
                if type(msg) == unicode:
                    msg = str(msg)

            repeat = 1
            total_sent = 0
            future = time.time() + (timeout or self.timeout)
            while repeat:
                repeat = 0
                while True:
                    # Attempt to send all.
                    # This won't work if the network buffer is already full.
                    try:
                        self.debug_print("Attempting to send: ")
                        self.debug_print(
                                str(len(msg[total_sent:self.chunk_size])))
                        self.debug_print(
                                "Blocking mode = " + str(self.s.gettimeout()))
                        bytes_sent = self.s.send(
                                msg[total_sent:self.chunk_size])
                    except socket.timeout as e:
                        err = e.args[0]
                        self.debug_print("Con send: " + str(e))
                        self.debug_print(err)
                        if err == "timed out":
                            return 0
                    except socket.error as e:
                        err = e.args[0]
                        self.debug_print("Con send: " + str(e))
                        if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                            break
                        else:
                            # Connection closed or other problem.
                            self.debug_print("Con send closing other")
                            self.close()
                            return 0

                    # Connection broken.
                    if not bytes_sent or bytes_sent is None:
                        self.debug_print(bytes_sent)
                        self.debug_print("Con send bytes none!")
                        self.close()
                        return 0

                    # How much has been sent?
                    total_sent += bytes_sent

                    # Avoid looping forever.
                    if self.blocking and not send_all:
                        break

                    # Everything sent.
                    if total_sent >= len(msg):
                        break

                    # Don't block.
                    if not send_all:
                        break

                    # Avoid 100% CPU.
                    time.sleep(0.002)

                # Avoid looping forever.
                if time.time() >= future:
                    repeat = 0
                    break

                # Send the rest if blocking:
                if total_sent < len(msg) and send_all:
                    repeat = 1

            self.alive = time.time()
            return total_sent
        except Exception as e:
            self.debug_print("Con send: " + str(e))
            error = parse_exception(e)
            log_exception(error_log_path, error)
            self.close()

    # Blocking or non-blocking.
    def recv(self, n, encoding="unicode", timeout=5):
        # Sanity checking.
        assert n

        # Disconnect.
        if not self.connected:
            if encoding == "unicode":
                return u""
            else:
                return b""

        # Update timeout.
        if timeout != self.timeout and self.blocking:
            self.set_blocking(self.blocking, timeout)

        # Check socket is in correct blocking mode.
        if self.debug:
            blocking = self.s.gettimeout()
            if self.blocking:
                assert (blocking >= 1 or blocking is None)
            else:
                assert (blocking == 0.0)

        try:
            # Save current buffer state.
            temp_buf = self.buf[:]

            # Clear buffer.
            self.buf = u""

            # Get data.
            future = time.time() + (timeout or self.timeout)
            while True:
                self.get_chunks(n, encoding=encoding)
                if not (len(self.buf) < n and self.connected and self.blocking):
                    break

                # Avoid 100% CPU.
                time.sleep(0.002)

                # Avoid looping forever.
                if time.time() >= future:
                    break

            # Save current buffer.
            ret = self.buf[:]

            # Restore old buffer.
            self.buf = temp_buf

            # Return results.
            if encoding != "unicode":
                # Convert from unicode string with latin-1 encoding
                # To a byte string.
                if sys.version_info >= (3, 0, 0):
                    codes = []
                    for ch in ret:
                        codes.append(ord(ch))

                    if len(codes):
                        return bytes(codes)
                    else:
                        return b""
                else:
                    byte_str = b""
                    for ch in ret:
                        byte_str += chr(ord(ch))

                    return byte_str

            return ret
        except Exception as e:
            self.debug_print("Recv closign e" + str(e))
            error = parse_exception(e)
            log_exception(error_log_path, error)
            self.close()
            if encoding == "unicode":
                return u""
            else:
                return b""

    # Sends a new message delimitered by a new line.
    # Blocking: blocks until entire line is sent for simplicity.
    def send_line(self, msg, timeout=5):
        # Sanity checking.
        assert (len(msg))

        # Not connected.
        if not self.connected:
            return 0

        # Update timeout.
        if timeout != self.timeout and self.blocking:
            self.set_blocking(self.blocking, timeout)

        # Check socket is in correct blocking mode.
        if self.debug:
            blocking = self.s.gettimeout()
            if self.blocking:
                assert (blocking >= 1 or blocking is None)
            else:
                assert (blocking == 0.0)

        try:
            # Convert to bytes Python 2 & 3
            if sys.version_info >= (3, 0, 0):
                if type(msg) == str:
                    msg = msg.encode("ascii")
            else:
                if type(msg) == unicode:
                    msg = str(msg)

            # Convert delimiter to bytes.
            if sys.version_info >= (3, 0, 0):
                msg += self.delimiter.encode("ascii")
            else:
                msg += str(self.delimiter)

            """
            The inclusion of the send_all flag makes this function behave like
            a blocking socket for the purposes of sending a full line even if
            the socket is non-blocking. It's assumed that lines will be small
            and if the network buffer is full this code won't end up as a
            bottleneck. (Otherwise you would have to check the number of bytes
            returned every time you sent a line which is quite annoying.)
            """
            ret = self.send(msg, send_all=1, timeout=timeout)

            return ret
        except Exception as e:
            self.debug_print("Send line closing" + str(e))
            error = parse_exception(e)
            log_exception(error_log_path, error)
            self.close()
            return 0

    # Receives a new message delimited by a new line.
    # Blocking or non-blocking.
    def recv_line(self, timeout=5):
        # Socket is disconnected.
        if not self.connected:
            return u""

        # Update timeout.
        if timeout != self.timeout and self.blocking:
            self.set_blocking(self.blocking, timeout)

        # Check socket is in correct blocking mode.
        if self.debug:
            blocking = self.s.gettimeout()
            if self.blocking:
                assert (blocking >= 1 or blocking is None)
            else:
                assert (blocking == 0.0)

        # Return existing reply.
        if len(self.replies):
            temp = self.replies[0]
            self.replies = self.replies[1:]
            return temp

        try:
            future = time.time() + (timeout or self.timeout)
            while True:
                self.update()

                # Socket is disconnected.
                if not self.connected:
                    return u""

                # Non-blocking.
                if not ((not len(self.replies) or len(
                        self.buf) >= self.max_buf) and self.blocking):
                    break

                # Timeout elapsed.
                if time.time() >= future and self.blocking:
                    break

                # Avoid 100% CPU.
                time.sleep(0.002)

            if len(self.replies):
                temp = self.replies[0]
                self.replies = self.replies[1:]
                return temp

            return u""
        except Exception as e:
            pass

    """
    These functions here make the class behave like a list. The
    list is a collection of replies received from the socket.
    Every iteration also has the bonus of checking for any
    new replies so it is very easy, for example to do:
    for replies in sock:
        To process replies without handling networking boilerplate.
    """

    def __len__(self):
        self.update()
        return len(self.replies)

    def __getitem__(self, key):
        self.update()
        return self.replies[key]

    def __setitem__(self, key, value):
        self.update()
        self.replies[key] = value

    def __delitem__(self, key):
        self.update()
        del self.replies[key]

    def pop_reply(self):
        # Get replies.
        replies = []
        for reply in self.replies:
            replies.append(reply)

        if len(replies):
            # Put replies back in the queue.
            self.replies = replies[1:]

            # Return the first reply.
            return replies[0]
        else:
            return None

    def __iter__(self):
        try:
            # Get replies.
            self.update()

            # Return replies.
            return iter(self.replies)
        finally:
            # Clear old replies.
            self.replies = []

    def __reversed__(self):
        return self.__iter__()


if __name__ == "__main__":
    """
    s = Sock("158.69.201.105", 8540)

    exit()
    s.send_line("SOURCE TCP")


    while 1:
        for reply in s:
            print(reply)

        time.sleep(0.5)


    # print(s.recv_line())
    # print("yes")
    """



import netifaces
import os
import struct
import sys
import time
import psutil
import gc
import ipaddress
import ntplib
import re
from future.moves.urllib.request import urlopen

import traceback
from .ipgetter import *
from .ip_routes import *

if platform.system() == "Linux":
    try:
        from pyroute2 import IPDB
        ip = IPDB()
    except IOError:
        ip = None
else:
    ip = None


def get_unused_port(port=None):
    """Checks if port is already in use."""
    if port is None or port < 1024 or port > 65535:
        port = random.randint(1024, 65535)
    assert(1024 <= port <= 65535)
    while True:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind(('', port))  # Try to open port
        except socket.error as e:
            if e.errno in (98, 10048):  # 98, 10048 means address already bound
                return get_unused_port(None)
            raise e
        s.close()
        return port


def log_exception(file_path, msg):
    msg = "\r\n" + msg
    with open(file_path, "a") as error_log:
        error_log.write(msg)


def parse_exception(e, output=0):
    tb = traceback.format_exc()
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    error = "%s %s %s %s %s" % (str(tb), str(exc_type), str(fname),
                                str(exc_tb.tb_lineno), str(e))

    if output:
        print(error)

    return str(error)


def ip2int(addr):
    return struct.unpack("!I", socket.inet_aton(addr))[0]


def int2ip(addr):
    return socket.inet_ntoa(struct.pack("!I", addr))

# Patches for urllib2 and requests to bind on specific interface.
# http://rossbates.com/2009/10/26/urllib2-with-multiple-network-interfaces/
true_socket = socket.socket


def build_bound_socket(source_ip):
    def bound_socket(*a, **k):
        if source_ip == "127.0.0.1":
            raise Exception("This function requires a LAN IP"
                            " (127.0.0.1 passed.)")

        sock = true_socket(*a, **k)
        sock.bind((source_ip, 0))
        return sock

    return bound_socket


def busy_wait(dt):
    # Use most accurate timer.
    if platform.system() == "Windows":
        timer = time.clock
    else:
        timer = time.time

    current_time = timer()
    while timer() < current_time + dt:
        pass


def get_ntp_worker(server):
    try:
        client = ntplib.NTPClient()
        response = client.request(server, version=3)
        ntp = response.tx_time
        return ntp
    except Exception as e:
        return None

ntp_servers = [
    "pool.ntp.org",
    "2.pool.ntp.org",
    "0.pool.ntp.org",
    "1.pool.ntp.org",
    "3.pool.ntp.org"
]

bad_ntp_servers = None


def remove_bad_ntp_servers():
    global ntp_servers
    global bad_ntp_servers

    bad_ntp_servers = []
    for server in ntp_servers:
        ntp = get_ntp_worker(server)
        if ntp is None:
            bad_ntp_servers.append(server)

    for server in bad_ntp_servers:
        ntp_servers.remove(server)


def get_ntp(local_time=0):
    global ntp_servers
    global bad_ntp_servers

    if local_time:
        return time.time()

    if bad_ntp_servers is None:
        remove_bad_ntp_servers()

    random.shuffle(ntp_servers, random.random)
    for server in ntp_servers:
        ntp = get_ntp_worker(server)
        if ntp is not None:
            return ntp

    return None


def get_default_gateway(interface="default"):
    if sys.version_info < (3, 0, 0):
        if type(interface) == str:
            interface = unicode(interface)
    else:
        if type(interface) == bytes:
            interface = interface.decode("utf-8")

    if platform.system() == "Windows":
        if interface == "default":
            default_routes = [r for r in get_ipv4_routing_table()
                              if r[0] == '0.0.0.0']
            if default_routes:
                return default_routes[0][2]

    try:
        gws = netifaces.gateways()
        if sys.version_info < (3, 0, 0):
            return gws[interface][netifaces.AF_INET][0].decode("utf-8")
        else:
            return gws[interface][netifaces.AF_INET][0]
    except:
        # This can also mean the machine is directly accessible.
        return None


def get_lan_ip(interface="default"):
    if sys.version_info < (3, 0, 0):
        if type(interface) == str:
            interface = unicode(interface)
    else:
        if type(interface) == bytes:
            interface = interface.decode("utf-8")

    # Get ID of interface that handles WAN stuff.
    default_gateway = get_default_gateway(interface)
    gateways = netifaces.gateways()
    wan_id = None
    if netifaces.AF_INET in gateways:
        gw_list = gateways[netifaces.AF_INET]
        for gw_info in gw_list:
            if gw_info[0] == default_gateway:
                wan_id = gw_info[1]
                break

    # Find LAN IP of interface for WAN stuff.
    interfaces = netifaces.interfaces()
    if wan_id in interfaces:
        families = netifaces.ifaddresses(wan_id)
        if netifaces.AF_INET in families:
            if_info_list = families[netifaces.AF_INET]
            for if_info in if_info_list:
                if "addr" in if_info:
                    return if_info["addr"]

    """
    Execution may reach here if the host is using
    virtual interfaces on Linux and there are no gateways
    which suggests the host is a VPS or server. In this
    case
    """
    if platform.system() == "Linux":
        if ip is not None:
            return ip.routes["8.8.8.8"]["prefsrc"]

    return None


def sequential_bind(n, interface="default"):
    # Get bind address.
    addr = ''
    if interface != "default":
        addr = get_lan_ip(interface)

    # Start the process.
    bound = 0
    mappings = []
    prospects = []
    while not bound:
        # Grab a random place to start.
        bound = 1
        start = random.randrange(1024, 65535 - n)

        # Use connect to see if its already bound.
        for i in range(0, n):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            local = start + i
            try:
                sock.connect((addr, local))
                sock.close()
                bound = 0
                break
            except socket.error:
                pass

            prospect = {
                "local": local,
            }
            prospects.append(prospect)

        for prospect in prospects:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                sock.bind((addr, prospect["local"]))
            except Exception as e:
                bound = 0
                for mapping in mappings:
                    mapping["sock"].close()
                mappings = []
                break

            mapping = {
                "source": prospect["local"],
                "sock": sock
            }
            mappings.append(mapping)

    return mappings


def is_port_forwarded(source_ip, port, proto, forwarding_servers):
    global true_socket
    if source_ip is not None:
        socket.socket = build_bound_socket(source_ip)

    ret = 0
    for forwarding_server in forwarding_servers:
        url = "http://" + forwarding_server["addr"] + ":"
        url += str(forwarding_server["port"])
        url += forwarding_server["url"]
        url += "?action=is_port_forwarded&port=" + str(port)
        url += "&proto=" + str(proto.upper())

        try:
            r = urlopen(url, timeout=2)
            response = r.read().decode("utf-8")
            if "yes" in response:
                ret = 1
                break
        except:
            continue

    socket.socket = true_socket
    return ret


def is_ip_private(ip_addr):
    if sys.version_info < (3, 0, 0):
        if type(ip_addr) == str:
            ip_addr = unicode(ip_addr)
    else:
        if type(ip_addr) == bytes:
            ip_addr = ip_addr.decode("utf-8")

    if ipaddress.ip_address(ip_addr).is_private and ip_addr != "127.0.0.1":
        return 1
    else:
        return 0


def is_ip_public(ip_addr):
    if sys.version_info < (3, 0, 0):
        if type(ip_addr) == str:
            ip_addr = unicode(ip_addr)
    else:
        if type(ip_addr) == bytes:
            ip_addr = ip_addr.decode("utf-8")

    if is_ip_private(ip_addr):
        return 0
    elif ip_addr == "127.0.0.1":
        return 0
    else:
        return 1


def is_ip_valid(ip_addr):
    if sys.version_info < (3, 0, 0):
        if type(ip_addr) == str:
            ip_addr = unicode(ip_addr)
    else:
        if type(ip_addr) == bytes:
            ip_addr = ip_addr.decode("utf-8")

    try:
        ipaddress.ip_address(ip_addr)
        return 1
    except:
        return 0


def is_valid_port(port):
    try:
        port = int(port)
    except:
        return 0
    if port < 1 or port > 65535:
        return 0
    else:
        return 1


def memoize(function):
    memo = {}

    def wrapper(*args):
        if args in memo:
            return memo[args]
        else:
            rv = function(*args)
            memo[args] = rv
            return rv
    return wrapper


def extract_ip(s):
    ip = re.findall("[0-9]+[.][0-9]+[.][0-9]+[.][0-9]+", s)
    if len(ip):
        return ip[0]

    return ""


@memoize
def get_wan_ip(n=0):
    """
    That IP module sucks. Occasionally it returns an IP address behind
    cloudflare which probably happens when cloudflare tries to proxy your web
    request because it thinks you're trying to DoS. It's better if we just run
    our own infrastructure.
    """

    if n == 2:
        try:
            ip = myip()
            ip = extract_ip(ip)
            if is_ip_valid(ip):
                return ip
        except Exception as e:
            print(str(e))
            return None

    # Fail-safe: use centralized server for IP lookup.
    from pyp2p.net import forwarding_servers
    for forwarding_server in forwarding_servers:
        url = "http://" + forwarding_server["addr"] + ":"
        url += str(forwarding_server["port"])
        url += forwarding_server["url"]
        url += "?action=get_wan_ip"
        try:
            r = urlopen(url, timeout=5)
            response = r.read().decode("utf-8")
            response = extract_ip(response)
            if is_ip_valid(response):
                return response
        except Exception as e:
            print(str(e))
            continue

    time.sleep(1)
    return get_wan_ip(n + 1)


def request_priority_execution():
    gc.disable()
    sys.setcheckinterval(999999999)
    if sys.version_info > (3, 0, 0):
        sys.setswitchinterval(1000)
    p = psutil.Process(os.getpid())
    try:
        if platform.system() == "Windows":
            p.nice(psutil.HIGH_PRIORITY_CLASS)
        else:
            p.nice(10)
    except psutil.AccessDenied:
        pass

    return p


def release_priority_execution(p):
    sys.setcheckinterval(100)
    if sys.version_info > (3, 0, 0):
        sys.setswitchinterval(0.005)
    try:
        if platform.system() == "Windows":
            p.nice(psutil.NORMAL_PRIORITY_CLASS)
        else:
            p.nice(5)
    except psutil.AccessDenied:
        pass
    gc.enable()

if __name__ == "__main__":
    # print(get_wan_ip())
    # pass
    print("In lib")




import sys
import random
import requests
import binascii
import umsgpack
from ast import literal_eval
from future.moves.urllib.parse import urlencode
#from multiprocessing import Process as Thread, Event
from threading import Thread, Event
from storjkademlia.node import Node as KadNode
from pyp2p.lib import is_ip_valid, is_valid_port
from twisted.internet import defer

import json
import string
import binascii

try:
    from Queue import Queue  # py2
except ImportError:
    from queue import Queue  # py3

import time
import logging

dht_msg_endpoint = "http://185.61.148.22/dht_msg.php"
# dht_msg_endpoint = "http://localhost/dht_msg.php"
logging.basicConfig()
log = logging.getLogger(__name__)

LONG_POLLING = True
RESERVATION_TIMEOUT = (10 * 60) - 5
MUTEX_TIMEOUT = RESERVATION_TIMEOUT
ALIVE_TIMEOUT = 60 - 5


class DHTProtocol:
    def __init__(self):
        self.messages_received = None


class DHT:
    def __init__(self, node_id=None, ip=None, port=0, password=None, network_id="default", debug=0, networking=1):
        self.node_id = node_id or self.rand_str(20)
        if sys.version_info >= (3, 0, 0):
            if type(self.node_id) == str:
                self.node_id = self.node_id.encode("ascii")
        else:
            if type(self.node_id) == unicode:
                self.node_id = str(self.node_id)

        self.node_id = binascii.hexlify(self.node_id).decode('utf-8')
        self.password = password or self.rand_str(30)
        self.ip = ip
        self.port = port
        self.network_id = network_id
        self.check_interval = 3  # For slow connections, unfortunately.
        self.last_check = 0
        self.debug = debug
        self.networking = networking
        self.relay_links = {}
        self.protocol = DHTProtocol()
        self.is_registered = Event()
        self.is_mutex_ready = Event()
        self.is_neighbours_ready = Event()
        self.handles = []
        self.threads = []
        self.running = 1
        self.has_mutex = 0
        self.neighbours = []

        # Register a new "account."
        if self.networking:
            self.register(self.node_id, self.password)
            self.is_registered.wait(5)
            self.mutex_loop()
            self.is_mutex_ready.wait(5)
            self.alive_loop()
            self.find_neighbours_loop()
            self.is_neighbours_ready.wait(5)
            assert(self.is_mutex_ready.is_set())
            assert(self.is_registered.is_set())

        self.message_handlers = set()

    def stop(self):
        self.running = 0
        for handle in self.handles:
            handle.close()
            # handle.raw._fp.close()

    def hook_queue(self, q):
        self.protocol.messages_received = q
        self.check_for_new_messages()

    def retry_in_thread(self, f, args={"args": None}, check_interval=2):
        def thread_loop(this_obj):
            while 1:
                try:
                    while not f(**args) and this_obj.running:
                        time.sleep(check_interval)

                        if not this_obj.running:
                            return

                    return
                except Exception as e:
                    print("unknown exception")
                    print(e)
                    time.sleep(1)

        t = Thread(target=thread_loop, args=(self,))
        t.setDaemon(True)
        self.threads.append(t)
        t.start()

        return t

    def check_for_new_messages(self):
        def do(args):
            for msg in self.list(self.node_id, self.password):
                self.protocol.messages_received.put(msg)

            return 0

        if LONG_POLLING:
            self.retry_in_thread(do, check_interval=0.1)
        else:
            self.retry_in_thread(do, check_interval=2)

    def mutex_loop(self):
        def do(args):
            # Requests a mutex from the server.
            call = dht_msg_endpoint + "?call=get_mutex&"
            call += urlencode({"node_id": self.node_id}) + "&"
            call += urlencode({"password": self.password})

            # Make API call.
            ret = requests.get(call, timeout=5).text
            if "1" in ret or "2" in ret:
                self.has_mutex = int(ret)
            self.is_mutex_ready.set()

            return 0

        self.retry_in_thread(do, check_interval=MUTEX_TIMEOUT)

    def alive_loop(self):
        def do(args):
            # Requests a mutex from the server.
            call = dht_msg_endpoint + "?call=last_alive&"
            call += urlencode({"node_id": self.node_id}) + "&"
            call += urlencode({"password": self.password})

            # Make API call.
            ret = requests.get(call, timeout=5)

            return 0

        self.retry_in_thread(do, check_interval=ALIVE_TIMEOUT)

    def can_test_knode(self, id):
        for neighbour in self.neighbours:
            if neighbour.id == id:
                if neighbour.can_test:
                    return 1

        return 0

    def has_testable_neighbours(self):
        for neighbour in self.neighbours:
            if neighbour.can_test:
                return 1

        return 0

    def find_neighbours_loop(self):
        def do(args):
            # Requests a mutex from the server.
            call = dht_msg_endpoint + "?call=find_neighbours&"
            call += urlencode({"node_id": self.node_id}) + "&"
            call += urlencode({"password": self.password}) + "&"
            call += urlencode({"network_id": self.network_id})

            # Make API call.
            ret = requests.get(call, timeout=5).text
            ret = json.loads(ret)
            if type(ret) == dict:
                ret = [ret]

            # Convert to kademlia neighbours.
            neighbours = []
            for neighbour in ret:
                if not is_ip_valid(neighbour["ip"]):
                    continue

                neighbour["port"] = int(neighbour["port"])
                if not is_valid_port(neighbour["port"]):
                    continue

                knode = KadNode(
                    id=binascii.unhexlify(neighbour["id"].encode("ascii")),
                    ip=neighbour["ip"],
                    port=neighbour["port"],
                    can_test=int(neighbour["can_test"])
                )

                neighbours.append(knode)

            self.neighbours = neighbours
            self.is_neighbours_ready.set()

            return 0

        self.retry_in_thread(do, check_interval=ALIVE_TIMEOUT)

    def get_neighbours(self):
        return self.neighbours

    def add_relay_link(self, dht):
        node_id = binascii.hexlify(dht.get_id())
        self.relay_links[node_id.decode("utf-8")] = dht

    def debug_print(self, msg):
        if self.debug:
            print(str(msg))

    def add_message_handler(self, handler):
        self.message_handlers.add(handler)

    def remove_transfer_request_handler(self, handler):
        pass

    def rand_str(self, length):
        return ''.join(random.choice(string.digits + string.ascii_lowercase +
                                     string.ascii_uppercase
                                     ) for i in range(length))

    def register(self, node_id, password):
        def do(node_id, password):
            try:
                # Registers a new node to receive messages.
                call = dht_msg_endpoint + "?call=register&"
                call += urlencode({"node_id": node_id}) + "&"
                call += urlencode({"password": password}) + "&"
                call += urlencode({"port": self.port}) + "&"
                call += urlencode({"network_id": self.network_id})
                if self.ip is not None:
                   call += "&" + urlencode({"ip": self.ip})

                # Make API call.
                ret = requests.get(call, timeout=5)
                self.handles.append(ret)
                if "success" not in ret.text:
                    return 0
                self.is_registered.set()
                return 1
            except Exception as e:
                print(e)
                self.debug_print("Register timed out in DHT msg")

            self.debug_print("DHT REGISTER FAILED")
            return 0

        mappings = {
            "node_id": node_id,
            "password": password
        }
        self.retry_in_thread(do, mappings)

    def build_dht_response(self, msg):
        msg = binascii.unhexlify(msg)
        msg = umsgpack.unpackb(msg)
        try:
            str_types = [type(u""), type(b"")]
            if type(msg) in str_types:
                msg = literal_eval(msg)
        except:
            msg = str(msg)

        return msg

    def serialize_message(self, msg):
        msg = umsgpack.packb(msg)
        msg = binascii.hexlify(msg)
        return msg

    def async_dht_put(self, key, value):
        d = defer.Deferred()
        def do(args):
            t = self.put(key, value, list_pop=0)
            while t.isAlive():
                time.sleep(1)

            d.callback("success")
            return 1

        self.retry_in_thread(do)
        return d

    def async_dht_get(self, key):
        d = defer.Deferred()
        def do(args):
            ret = self.list(node_id=key, list_pop=0, timeout=5)
            if len(ret):
                d.callback(ret[0])
            else:
                d.callback(None)
            return 1

        self.retry_in_thread(do)
        return d

    def put(self, node_id, msg, list_pop=1):
        def do(node_id, msg):
            if node_id in self.relay_links:
                relay_link = self.relay_links[node_id]
                msg = self.build_dht_response(self.serialize_message(msg))
                relay_link.protocol.messages_received.put_nowait(msg)
                return 1

            try:
                # Send a message directly to a node in the "DHT"
                call = dht_msg_endpoint + "?call=put&"
                call += urlencode({"dest_node_id": node_id}) + "&"
                msg = self.serialize_message(msg)
                call += urlencode({"msg": msg}) + "&"
                call += urlencode({"node_id": self.node_id}) + "&"
                call += urlencode({"password": self.password}) + "&"
                call += urlencode({"list_pop": list_pop})

                # Make API call.
                ret = requests.get(call, timeout=5)
                self.handles.append(ret)
                if "success" not in ret.text:
                    return 0

                return 1

            except Exception as e:
                # Reschedule call.
                self.debug_print("DHT PUT TIMED OUT")
                self.debug_print(e)
                self.debug_print("Rescheduling DHT PUT")

            self.debug_print("PUT FAILED")
            return 0

        mappings = {
            "node_id": node_id,
            "msg": msg
        }
        return self.retry_in_thread(do, mappings)

    def list(self, node_id=None, password=None, list_pop=1, timeout=None):
        if not self.networking:
            return []

        node_id = node_id or self.node_id
        password = password or self.password
        try:
            # Get messages send to us in the "DHT"
            call = dht_msg_endpoint + "?call=list&"
            call += urlencode({"node_id": node_id}) + "&"
            call += urlencode({"password": password}) + "&"
            call += urlencode({"list_pop": list_pop})

            # Make API call.
            if timeout is None:
                if LONG_POLLING:
                    timeout = None
                else:
                    timeout = 4
            ret = requests.get(call, timeout=timeout)
            self.handles.append(ret)
            content_gen = ret.iter_content()
            messages = ret.text
            messages = json.loads(messages)

            # List.
            if type(messages) == dict:
                messages = [messages]

            # Return a list of responses.
            ret = []
            if type(messages) == list:
                for msg in messages:
                    dht_response = self.build_dht_response(msg)
                    ret.append(dht_response)

            return ret
        except Exception as e:
            print("EXCEPTION IN DHT MSG LIST")
            self.debug_print("Exception in dht msg list")
            print(e)
            return []

    def direct_message(self, node_id, msg):
        return self.send_direct_message(node_id, msg)

    def relay_message(self, node_id, msg):
        return self.send_direct_message(node_id, msg)

    def repeat_relay_message(self, node_id, msg):
        return self.send_direct_message(node_id, msg)

    def async_direct_message(self, node_id, msg):
        return self.send_direct_message(node_id, msg)

    def send_direct_message(self, node_id, msg):
        if sys.version_info >= (3, 0, 0):
            if type(node_id) == bytes:
                node_id = binascii.hexlify(node_id).decode("utf-8")
        else:
            if type(node_id) == str:
                node_id = binascii.hexlify(node_id).decode("utf-8")

        if type(node_id) != str:
            node_id = node_id.decode("utf-8")

        self.put(node_id, msg)

    def get_id(self):
        node_id = self.node_id
        if sys.version_info >= (3, 0, 0):
            if type(node_id) == str:
                node_id = node_id.encode("ascii")
        else:
            if type(node_id) == unicode:
                node_id = str(node_id)

        return binascii.unhexlify(node_id)

    def has_messages(self):
        return not self.protocol.messages_received.empty()

    def get_messages(self):
        result = []
        if self.has_messages():
            while not self.protocol.messages_received.empty():
                result.append(self.protocol.messages_received.get())

            # Run handlers on messages.
            old_handlers = set()
            for received in result:
                for handler in self.message_handlers:
                    expiry = handler(
                        self,
                        received
                    )

                    if expiry == -1:
                        old_handlers.add(handler)

            # Expire old handlers.
            for handler in old_handlers:
                self.message_handlers.remove(handler)

            return result

        return result


if __name__ == "__main__":
    #node1 = DHT()
    #print(node1.get_id())
    #print(node1.node_id)

    pass
    """
    node1 = DHT()
    node2 = DHT()
    node1.put(node2.node_id, "test")
    running = 1
    time.sleep(5)
    node1.stop()
    node2.stop()
    """

    """
    #print(node2.protocol.messages_received.get())
    #print(node2.get_messages())


    while not node2.has_messages() and running:
        for msg in node2.get_messages():
            running = 0
            print(msg)

    print("No longer runnig")
    """


    """
    #dht_node = DHT(node_id=b"\111" * 20, password="svymQQzF1j7FGmYf8fENs4mvRd")

    dht_node = DHT(node_id=u"T", password="svymQQzF1j7FGmYf8fENs4mvRd")

    x = [("a", 2), ("b!%--", 2)]
    dht_node.put(dht_node.node_id, x)
    print(dht_node.list(dht_node.node_id, dht_node.password))



    exit()

    print(dht_node.node_id)
    print(dht_node.get_id())
    print(type(dht_node.get_id()))


    dht_node.send_direct_message(dht_node.node_id, u"test")
    print(dht_node.list(dht_node.node_id, dht_node.password))
    exit()



    print(dht_node.node_id)
    print(dht_node.password)
    print(dht_node.list(dht_node.node_id, dht_node.password))
    """



import select

from .lib import *
from .sock import *

from future.moves.urllib.request import urlopen, Request

import platform
import subprocess
from threading import Thread

"""
This is a modified version of "UPnP-Exploiter." It's been
changed to work with Python 3.3, the code has been
restructured to make it modular, commented, and easier to use.
I've also removed the original exploit code and only the port
forwarding code remains.

Original code available here: https://github.com/dc414/Upnp-Exploiter
Credits to "Anarchy Angel", "Ngharo", and www.dc414.org
for the code.
"""


class UPnP:
    def __init__(self, interface=u"default"):
        """
        Port used to listen to UPnP replies on.
        This port is actually arbitrary because
        all the sockets used are datagram sockets
        bound to all address which lets
        the socket hear broadcasts.
        """
        self.listen_port = 49170

        # Port that UPnP configured hosts listen on.
        self.upnp_port = 1900

        # Address used for IPv4 multicasts.
        self.multicast = b"239.255.255.250"

        # Number of seconds to wait for replies.
        self.reply_wait = 3

        # Socket timeout.
        self.timeout = 2

        # Networking interface.
        self.interface = interface

        # Address of a UPnP gateway.
        self.gateway = []

    # Uses broadcasting to find default UPnP compatible gateway.
    def find_gateway(self):
        replies = []

        # Create socket for UDP broadcasts.
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(('', self.upnp_port))  # All addresses.
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        s.setblocking(0)

        # Broadcast search message to multicast address.
        search_msg = b"M-SEARCH * HTTP/1.1\r\n"
        search_msg += b"HOST: %s" + self.multicast + b":"
        if sys.version_info >= (3, 0, 0):
            search_msg += str(self.upnp_port).encode("ascii")
        else:
            search_msg += str(self.upnp_port)
        search_msg += b"\r\nST: ssdp:all\r\n"
        search_msg += b"""MAN: "ssdp:discover"\r\n"""
        search_msg += b"MX: 1\r\n"
        search_msg += b"\r\n"
        s.sendto(search_msg, (self.multicast, self.upnp_port))

        # Receive replies for n seconds..
        old_time = time.time()
        while (int(time.time()) - int(old_time)) < self.reply_wait:
            res = select.select([s], [], [], self.timeout)
            if len(res[0]):
                (string, addr) = res[0][0].recvfrom(1024)
                replies.append([addr[0], string])

        # Cleanup socket.
        if s is not None:
            s.close()
            s = None

        # Error: no UPnP replies - try guess gateway.
        if not replies:
            default_gateway = get_default_gateway(self.interface)
            if default_gateway is None or default_gateway == {}:
                return None
            else:
                # Optimise scanning.
                likely_candidates = [80, 1780, 1900, 1981, 2468, 5555, 5678,
                                     49000, 55345, 65535]

                def check_gateway(self, ip, port):
                    try:
                        # Fast connect() / SYN open scanning.
                        s = Sock(ip, port, blocking=1, timeout=5,
                                 interface=self.interface)
                        s.close()

                        # Build http request.
                        gateway_addr = "http://" + str(ip) + ":" +\
                                       str(port) + "/"
                        buf = urlopen(gateway_addr, timeout=self.timeout).\
                            read().decode("utf-8")

                        # Check response is XML and device is a router.
                        if 'InternetGatewayDevice' in buf:
                            return self.gateway.append(gateway_addr)
                    except:
                        return

                # Brute force port by scanning.
                for port in likely_candidates:
                    t = Thread(target=check_gateway,
                               args=(self, default_gateway, port))
                    t.start()

                time.sleep(5)
                if len(self.gateway):
                    return self.gateway[0]
                else:
                    return None

        # Find gateway address in replies.
        gateway_addr = None
        pdata = list(dict((x[0], x) for x in replies).values())
        # return None
        rh = []
        for L in pdata:
            rh.append(L[0])
        hosts = []
        pd = []
        for host in rh:
            try:
                spot = rh.index(host)
                hdata = pdata[spot][1]
                url = 'http://' + host + ':'
                port = re.findall("http:\/\/[0-9\.]+:(\d.+)",
                                  hdata.decode("utf-8"))
                url += port[0]
                p = urlopen(url, timeout=self.timeout)
                rd = re.findall('schemas-upnp-org:device:([^:]+)',
                                p.read().decode("utf-8"))
                if rd[0] == 'InternetGatewayDevice':
                    gateway_addr = url
                    break
            except:
                continue

        return gateway_addr

    def forward_port(self, proto, src_port, dest_ip, dest_port=None):
        """
        Creates a new mapping for the default gateway to forward ports.
        Source port is from the perspective of the original client.
        For example, if a client tries to connect to us on port 80,
        the source port is port 80. The destination port isn't
        necessarily 80, however. We might wish to run our web server
        on a different port so we can have the router forward requests
        for port 80 to another port (what I call the destination port.)

        If the destination port isn't specified, it defaults to the
        source port. Proto is either TCP or UDP. Function returns None
        on success, otherwise it raises an exception.
        """

        proto = proto.upper()
        valid_protos = ["TCP", "UDP"]
        if proto not in valid_protos:
            raise Exception("Invalid protocol for forwarding.")

        valid_ports = range(1, 65535)
        if src_port not in valid_ports:
            raise Exception("Invalid port for forwarding.")

        # Source port is forwarded to same destination port number.
        if dest_port is None:
            dest_port = src_port

        # Use UPnP binary for forwarding on Windows.
        if platform.system() == "Windows":
            cmd = "upnpc-static.exe -a %s %s %s %s" % (get_lan_ip(),
                                                       str(src_port),
                                                       str(dest_port),
                                                       proto)
            out, err = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE).communicate()
            if "is not recognized" in err:
                raise Exception("Missing upnpc-static.exe")
            
            return

        # Find gateway address.
        gateway_addr = self.find_gateway()
        if gateway_addr is None:
            raise Exception("Unable to find UPnP compatible gateway.")

        # Get control URL.
        rhost = re.findall('([^/]+)', gateway_addr)
        res = urlopen(gateway_addr, timeout=self.timeout).read().decode("utf-8")
        res = res.replace('\r', '')
        res = res.replace('\n', '')
        res = res.replace('\t', '')
        pres = res.split('<serviceId>urn:upnp-org:serviceId:WANIPConn1'
                         '</serviceId>')
        p2res = pres[1].split('</controlURL>')
        p3res = p2res[0].split('<controlURL>')
        ctrl = p3res[1]
        rip = res.split('<presentationURL>')
        rip1 = rip[1].split('</presentationURL>')
        router_ip = rip1[0]

        port_map_desc = "PyP2P"
        msg = \
            '<?xml version="1.0"?><s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"><s:Body><u:AddPortMapping xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1"><NewRemoteHost></NewRemoteHost><NewExternalPort>' \
            + str(src_port) \
            + '</NewExternalPort><NewProtocol>' + str(proto) + '</NewProtocol><NewInternalPort>' \
            + str(dest_port) + '</NewInternalPort><NewInternalClient>' + str(dest_ip) \
            + '</NewInternalClient><NewEnabled>1</NewEnabled><NewPortMappingDescription>' + str(port_map_desc) + '</NewPortMappingDescription><NewLeaseDuration>0</NewLeaseDuration></u:AddPortMapping></s:Body></s:Envelope>'

        # Attempt to add new port map.
        x = 'http://' + rhost[1] + '/' + ctrl
        if sys.version_info >= (3, 0, 0):
            msg = bytes(msg, "utf-8")

        req = Request('http://' + rhost[1] + '/' + ctrl, msg)
        req.add_header('SOAPAction',
                       '"urn:schemas-upnp-org:service:WANIPConnection:1#AddPortMapping"'
                       )
        req.add_header('Content-type', 'application/xml')
        res = urlopen(req, timeout=self.timeout)


if __name__ == "__main__":
    """
    if sys.version_info < (3,0,0):


    port = 51020
    addr = "192.168.0.60"
    forwarding_servers = [{"addr": "158.69.201.105", "port": 80, "url": "/pyp2p/net.php"}]


    print(UPnP().forward_port("TCP", port, addr))
    print(is_port_forwarded(get_lan_ip(), str(port), "TCP", forwarding_servers))
    """



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    version='0.7.8',
    name='pyp2p',
    description='Python P2P networking library',
    keywords=('NAT traversal, TCP hole punching, simultaneous open, UPnP,'
              ' NATPMP, P2P, Peer-to-peer networking library, python'),
    long_description=long_description,
    url='http://github.com/Storj/pyp2p',
    author='Storj',
    author_email='matthew@storj.io',
    test_suite="tests",
    license='MIT',
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'netifaces>=0.10.4',
        'ntplib>=0.3.3',
        'twisted>=15.4.0',
        'ipaddress>=1.0.14',
        'requests>=2.8.1',
        'pyroute2>=0.3.15'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



"""
Given an NTP accurate time, this module computes an
approximation of how far off the system clock is
from the NTP time (clock skew.) The algorithm was
taken from gtk-gnutella.

https://github.com/gtk-gnutella/gtk-gnutella/
blob/devel/src/core/clock.c 
"""


from decimal import Decimal

from .lib import *


class SysClock:
    def __init__(self, clock_skew=Decimal("0")):
        self.enough_data = 40
        self.min_data = 20
        self.max_sdev = 60
        self.clean_steps = 3
        self.data_points = []
        self.clock_skew = clock_skew
        if not self.clock_skew:
            self.collect_data_points()
            self.clock_skew = self.calculate_clock_skew()

    def time(self):
        return Decimal(time.time()) - self.clock_skew

    def collect_data_points(self):
        while len(self.data_points) < self.enough_data + 10:
            clock_skew = Decimal(time.time()) - Decimal(get_ntp())
            self.data_points.append(clock_skew)

    def statx_n(self, data_points):
        return len(data_points)

    def statx_avg(self, data_points):
        total = Decimal("0")
        n = self.statx_n(data_points)
        for i in range(0, n):
            total += data_points[i]

        return total / Decimal(n)

    def statx_sdev(self, data_points):
        def _ss(data):
            """Return sum of square deviations of sequence data."""
            c = self.statx_avg(data)
            ss = sum((x-c)**2 for x in data)
            return ss

        def pstdev(data):
            """Calculates the population standard deviation."""
            n = len(data)
            if n < 2:
                raise ValueError('variance requires at least two data points')
            ss = _ss(data)
            pvar = ss/n  # the population variance
            return pvar**Decimal("0.5")

        return pstdev(data_points)

    def calculate_clock_skew(self):
        """
        Computer average and standard deviation
        using all the data points.
        """
        n = self.statx_n(self.data_points)

        """
        Required to be able to compute the standard
        deviation.
        """
        if n < 1:
            return Decimal("0")

        avg = self.statx_avg(self.data_points)
        sdev = self.statx_sdev(self.data_points)

        """
        Incrementally remove aberration points.
        """
        for k in range(0, self.clean_steps):
            """
            Remove aberration points: keep only
            the sigma range around the average.
            """
            min_val = avg - sdev
            max_val = avg + sdev

            cleaned_data_points = []
            for i in range(0, n):
                v = self.data_points[i]
                if v < min_val or v > max_val:
                    continue
                cleaned_data_points.append(v)

            self.data_points = cleaned_data_points[:]

            """
            Recompute the new average using the
            "sound" points we kept.
            """
            n = self.statx_n(self.data_points)

            """
            Not enough data to compute standard
            deviation.
            """
            if n < 2:
                break

            avg = self.statx_avg(self.data_points)
            sdev = self.statx_sdev(self.data_points)

            if sdev <= self.max_sdev or n < self.min_data:
                break

        """
        If standard deviation is too large still, we
        cannot update our clock. Collect more points.

        If we don't have a minimum amount of data,
        don't attempt the update yet, continue collecting.
        """
        if sdev > self.max_sdev or n < self.min_data:
            return Decimal("0")

        return avg
        
if __name__ == "__main__":
    sys_clock = SysClock()
    print(sys_clock.clock_skew)

    while 0:
        sys_clock = SysClock()
        ntp = Decimal(get_ntp())
        adjusted = sys_clock.time()
        dif = ntp - adjusted
        if dif < 0:
            dif = -dif

        print(dif)
        if dif < 0.05:
            break

    # print(get_ntp())
    # print(get_ntp())

    """
    print(sys_clock.time())
    print()
    print(get_ntp())
    print(sys_clock.time())
    print()
    print(get_ntp())
    print(sys_clock.time())
    """

